from . augmentation import SingleDigitAugmentation
from . augmentation import SequenceAugmentation
from . shearing import RandomImageWidthChange
from . shearing import AscendingImageWidthChange
from . noise import GaussianNoise
from . noise import SPNoise
from . noise import PoissonNoise
from . noise import SpeckleNoise
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__all__ = [
    'SingleDigitAugmentation',
    'GaussianNoise',
    'SpeckleNoise',
    'PoissonNoise',
    'SPNoise',
    "SequenceAugmentation",
    "RandomImageWidthChange",
    "AscendingImageWidthChange"
]



import unittest
import numpy as np
from digitgen import DigitGenerator
from digitgen import test_annotations


class TestSpaceBetweenDigits(unittest.TestCase):
    def test_normal(self):
        digit_gen = DigitGenerator(10, samples=10, image_size=(512, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == 10)

    def test_constant_space_between_digits(self):
        digit_gen = DigitGenerator(10, samples=20, image_size=(512, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        digit_gen.add_width_digits(pixel_value=5)
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == 10)

    def test_random_space_between_digits(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(512, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        digit_gen.add_width_digits(mode="random", pixel_range=[0, 10])
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == 10)



from abc import ABC, abstractmethod
from random import random


class Augmentation(ABC):
    def __init__(self, probability) -> None:
        super().__init__()
        self.probability = probability

    @abstractmethod
    def augment(self, images, annotations):
        pass

    def apply_augmentation(self, images, annotations):
        value = random()
        if value <= self.probability:
            return self.augment(images, annotations)

        return images, annotations


class SingleDigitAugmentation(Augmentation):
    def __init__(self, probability) -> None:
        super().__init__(probability)

    @abstractmethod
    def augment(self, image, annotation):
        pass


class SequenceAugmentation(Augmentation):
    def __init__(self, probability) -> None:
        super().__init__(probability)

    @abstractmethod
    def augment(self, image, annotation):
        pass



import random
import cv2 as cv

from .augmentation import SequenceAugmentation


class AscendingImageWidthChange(SequenceAugmentation):
    def __init__(self, width_increment=2, range_type="pixels"):
        super(AscendingImageWidthChange, self).__init__(probability=1)
        self.width_increment = width_increment
        self.range_type = range_type

    def augment(self, images, annotations):
        index_range = range(len(images))
        shift_factor = 0
        add_factor = 0

        if self.range_type == "pixels":
            for __id in index_range:
                img, ann = images[__id], annotations[__id]
                images[__id] = cv.resize(img, dsize=(img.shape[1] + add_factor, img.shape[0]))

                ann["bbox"][0] += shift_factor
                shift_factor += add_factor
                ann["bbox"][2] += shift_factor

                add_factor += self.width_increment


        elif self.range_type == "percentage":
            for __id in index_range:
                img, ann = images[__id], annotations[__id]
                rand_int = int(img.shape[1] * self.width_increment)

                images[__id] = cv.resize(img, dsize=(img.shape[1] + add_factor, img.shape[0]))

                ann["bbox"][0] += shift_factor
                shift_factor += add_factor
                ann["bbox"][2] += shift_factor

                add_factor += rand_int

        return images, annotations


class RandomImageWidthChange(SequenceAugmentation):
    def __init__(self, width_range=None, range_type="pixels"):
        super(RandomImageWidthChange, self).__init__(probability=1)
        if width_range is None:
            width_range = [0, 10]
        self.width_range = width_range
        self.range_type = range_type

    def augment(self, images, annotations):
        if self.range_type == "pixels":
            shift_factor = 0
            for __id in range(len(images)):
                rand_int = random.randint(self.width_range[0], self.width_range[1])

                img, ann = images[__id], annotations[__id]
                images[__id] = cv.resize(img, dsize=(img.shape[1] + rand_int, img.shape[0]))

                ann["bbox"][0] += shift_factor
                shift_factor += rand_int
                ann["bbox"][2] += shift_factor

        elif self.range_type == "percentage":
            shift_factor = 0
            for __id in range(len(images)):
                rand_int = random.randint(self.width_range[0], self.width_range[1])

                img, ann = images[__id], annotations[__id]
                rand_int = img.shape[1] * rand_int // 100

                images[__id] = cv.resize(img, dsize=(img.shape[1] + rand_int, img.shape[0]))

                ann["bbox"][0] += shift_factor
                shift_factor += rand_int
                ann["bbox"][2] += shift_factor

        return images, annotations



import unittest
import numpy as np
from digitgen import DigitGenerator
from digitgen import test_annotations


class TestChangeCommonConfig(unittest.TestCase):
    def test_default_transpose_generation(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])

        self.assertTrue(ret_arr.shape == (100, 48, 128, 3))

        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.change_common_config({"background_color": (0, 0, 0), "digit_color": (255, 255, 255)})
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])

        self.assertTrue(ret_arr.shape == (100, 48, 128, 3))

    def test_invert_color_generation(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])

        self.assertTrue(ret_arr.shape == (100, 48, 128, 3))

        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        digit_gen.randomly_invert_color()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])

        self.assertTrue(ret_arr.shape == (100, 48, 128, 3))


import numpy as np

from .digit import DigitConfig, DigitSequence
from .utils import CONFIGURATION
from .utils import convert_to_grayscale
from .font import FontConfig
from .utils import (format_annotations, generate_random_digits, generate_random_digits_with_probability, add_spaces,
                    category_id, generate_random_digits_with_positional_probability, change_between_configs,
                    randomly_invert_config_color)

from .augmentation import SequenceAugmentation, SingleDigitAugmentation


class DigitGenerator(object):
    def __init__(self, digit_size: int,
                 font_type: str = "terminal-grotesque-regular",
                 samples=1,
                 gray_scale=False,
                 transpose=False,
                 image_size=None,
                 augmentations=None) -> None:
        if augmentations is None:
            augmentations = []
        self.allowed_digits = None
        self.digit_size = digit_size
        self.samples = samples
        self.image_size = image_size
        self.digit_augmentations = []
        self.sequence_augmentation = []
        self.gray_scaled = gray_scale
        self.transpose = transpose

        for aug in augmentations:
            if isinstance(aug, SingleDigitAugmentation):
                self.digit_augmentations.append(aug)
            elif isinstance(aug, SequenceAugmentation):
                self.sequence_augmentation.append(aug)
            else:
                raise NotImplementedError

        self.font_type = FontConfig(font_type=font_type)
        self.config = CONFIGURATION
        self.config["common_configs"]["font_file_loc"] = self.font_type.get_font_file_location()

        self.generated_digits = None
        self.generated_configs = []
        self.memory = {}
        self.id2category = None
        self.category2id = None

    def change_common_config(self, changes: dict):
        for k, v in changes.items():
            self.config["common_configs"][k] = v

    def generate_digits(self, allowed_digits=None):
        if allowed_digits is None:
            allowed_digits = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
        random_lists = []
        for _ in range(self.samples):
            row = generate_random_digits(self.digit_size, allowed_digits)
            random_lists.append(row)

        self.allowed_digits = allowed_digits
        self.generated_digits = random_lists

    def generate_digits_with_choice_probability(self, probability: dict):
        random_lists = []
        for _ in range(self.samples):
            row = generate_random_digits_with_probability(self.digit_size, probability)
            random_lists.append(row)

        self.allowed_digits = list(probability.keys())
        self.generated_digits = random_lists

    def generate_digits_with_position_probability(self, probability: dict):
        random_lists = []
        for _ in range(self.samples):
            row = generate_random_digits_with_positional_probability(self.digit_size, probability)
            random_lists.append(row)

        all_values = []
        for __probability in probability.values():
            all_values += __probability.keys()

        self.allowed_digits = list(tuple(all_values))
        self.generated_digits = random_lists

    def generate_spaces(self, space_type="None", num_spaces=None, sectors=None, spaces_per_sector=None):
        assert self.generated_digits
        space_lists = []

        for row in self.generated_digits:
            row = add_spaces(row, space_type, num_spaces, sectors, spaces_per_sector, self.digit_size)
            space_lists.append(row)

        self.allowed_digits.append(" ")
        self.generated_digits = space_lists

    def generate_digit_config(self):
        id2category, category2id = category_id(self.allowed_digits)

        self.id2category = id2category
        self.category2id = category2id

        for row_id in range(self.samples):
            row = self.generated_digits[row_id]
            configurations = [DigitConfig.load_config(
                self.config, x, category2id[x]) for x in row]

            self.generated_configs.append(configurations)

    def add_width_digits(self, mode="constant", pixel_value=5, pixel_range=None):
        if pixel_range is None:
            pixel_range = [0, 20]

        change_between_configs(self.generated_configs, mode, pixel_value, pixel_range, self.samples,
                               self.digit_size)

    def randomly_invert_color(self, probability=0.5, background_color=(0, 0, 0), digit_color=(255, 255, 255)):
        randomly_invert_config_color(self.generated_configs, probability, background_color, digit_color)

    def generate_dataset(self):
        """
        Generator function of the dataset

        Returns:
            Sequence[tuple[np.array,dict]]: img_array,annotation
        """

        annotations = {"annotations": []}
        arrays = []
        for row in self.generated_configs:
            digits = DigitSequence(
                configs=row, size=self.image_size, memory=self.memory,
                sequence_augmentations=self.sequence_augmentation, digit_augmentations=self.digit_augmentations)
            array, annotation = digits.data()

            if self.transpose:
                array = np.transpose(array, axes=(1, 0, 2))

            arrays.append(array)
            format_annotations(annotations, annotation, self.id2category)

        arrays = np.array(arrays)
        if self.gray_scaled:
            arrays = convert_to_grayscale(arrays)
        return arrays, annotations



import unittest

import numpy as np

from digitgen import test_annotations
from digitgen.augmentation import GaussianNoise, SPNoise, SpeckleNoise, PoissonNoise, RandomImageWidthChange, \
    AscendingImageWidthChange
from digitgen.digit import DigitSequence, DigitConfig
from digitgen.font import FontConfig
from digitgen.utils import CONFIGURATION


class TestNoiseAugmentation(unittest.TestCase):
    def setUp(self) -> None:
        self.font_type = FontConfig()
        self.config = CONFIGURATION
        self.config["common_configs"]["font_file_loc"] = self.font_type.get_font_file_location()

    def test_gaussian_noise(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 758abc3847 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 75838iilco47 475845 5849"]
        digits = DigitSequence(configs=configurations, digit_augmentations=[GaussianNoise()])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_salt_pepper_noise(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 758vburb3847 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583847 475845 5849"]
        digits = DigitSequence(configs=configurations, digit_augmentations=[SPNoise()])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_speckle_noise(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583847 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583847 475845 5849"]
        digits = DigitSequence(configs=configurations, digit_augmentations=[SpeckleNoise()])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_poisson_noise(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583847 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583pkc 847 475845 5849"]
        digits = DigitSequence(configs=configurations, digit_augmentations=[PoissonNoise()])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_multiple_noise(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 75838ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "123 7583847 475845 5849"]
        digits = DigitSequence(configs=configurations,
                               digit_augmentations=[PoissonNoise(), GaussianNoise(), SPNoise(), SpeckleNoise()])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_shearing_augmentations(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations,
                               sequence_augmentations=[RandomImageWidthChange(width_range=[0, 20])])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations,
                               sequence_augmentations=[
                                   RandomImageWidthChange(width_range=[0, 99], range_type="percentage")])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)

    def test_ascending_shearing_augmentations(self):
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations)
        array, annotation = digits.data()
        test_annotations(array, annotation)

        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations,
                               sequence_augmentations=[AscendingImageWidthChange(width_increment=5)])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        configurations = [DigitConfig.load_config(self.config, x, 0) for x in "1237588ncjd 47 475845 5849"]
        digits = DigitSequence(configs=configurations,
                               sequence_augmentations=[
                                   AscendingImageWidthChange(width_increment=0.2, range_type="percentage")])
        array, annotation = digits.data()
        test_annotations(array, annotation)
        self.assertTrue(np.unique(array).shape[0] > 2)
